import sys


def apply_asyncio_patch():
    """
    Apply a patch to asyncio's proactor events on Windows when running Python 3.8 or 3.9.

    This patch is needed because of a bug in Python 3.8 and 3.9 that causes a
    RuntimeError to be raised when the event loop is closed while there are still
    open file descriptors. This bug was fixed in Python 3.10.

    The bug manifests itself when using the proactor event loop on Windows, which
    is the default event loop on Windows. The bug causes the event loop to be
    closed while there are still open file descriptors, which causes a RuntimeError
    to be raised.

    This patch catches the RuntimeError and ignores it, which allows the event loop
    to be closed properly.

    See https://bugs.python.org/issue39232 and https://github.com/python/cpython/issues/92841
    for more information.
    """

    if sys.platform == "win32" and (3, 8, 0) <= sys.version_info < (3, 11, 0):
        import asyncio.proactor_events as proactor_events

        original_del = proactor_events._ProactorBasePipeTransport.__del__

        def patched_del(self):
            try:
                original_del(self)
            except RuntimeError as e:
                if str(e) != "Event loop is closed":
                    raise

        proactor_events._ProactorBasePipeTransport.__del__ = patched_del


apply_asyncio_patch()



import base64
import json
import logging
import re
import shutil
import subprocess
from pathlib import Path
from typing import Optional
from urllib.parse import urlsplit, urlunsplit

import typer
from rich.console import Console

from safety.tool.constants import PUBLIC_REPOSITORY_URL, ORGANIZATION_REPOSITORY_URL
from safety.tool.resolver import get_unwrapped_command

from safety.console import main_console


logger = logging.getLogger(__name__)


class Pip:
    @classmethod
    def is_installed(cls) -> bool:
        """
        Checks if the PIP program is installed

        Returns:
            True if PIP is installed on system, or false otherwise
        """
        return shutil.which("pip") is not None

    @classmethod
    def configure_requirements(
        cls,
        file: Path,
        org_slug: Optional[str],
        console: Console = main_console,
    ) -> Optional[Path]:
        """
        Configures Safety index url for specified requirements file.

        Args:
            file (Path): Path to requirements.txt file.
            org_slug (str): Organization slug.
            console (Console): Console instance.
        """

        with open(file, "r+") as f:
            content = f.read()

            repository_url = (
                ORGANIZATION_REPOSITORY_URL.format(org_slug)
                if org_slug
                else PUBLIC_REPOSITORY_URL
            )
            index_config = f"-i {repository_url}\n"
            if content.find(index_config) == -1:
                f.seek(0)
                f.write(index_config + content)

                logger.info(f"Configured {file} file")
                return file
            else:
                logger.info(f"{file} is already configured. Skipping.")

        return None

    @classmethod
    def configure_system(
        cls, org_slug: Optional[str], console: Console = main_console
    ) -> Optional[Path]:
        """
        Configures PIP system to use to Safety index url.
        """
        try:
            repository_url = (
                ORGANIZATION_REPOSITORY_URL.format(org_slug)
                if org_slug
                else PUBLIC_REPOSITORY_URL
            )
            result = subprocess.run(
                [
                    get_unwrapped_command(name="pip"),
                    "config",
                    "set",
                    "global.index-url",
                    repository_url,
                ],
                capture_output=True,
            )

            if result.returncode != 0:
                logger.error(
                    f"Failed to configure PIP global settings: {result.stderr.decode('utf-8')}"
                )
                return None

            output = result.stdout.decode("utf-8")
            match = re.search(r"Writing to (.+)", output)

            if match:
                config_file_path = match.group(1).strip()
                return Path(config_file_path)

            logger.error("Failed to match the config file path written by pip.")
            return Path()
        except Exception:
            logger.exception("Failed to configure PIP global settings.")

        return None

    @classmethod
    def reset_system(cls, console: Console = main_console):
        # TODO: Move this logic and implement it in a more robust way
        try:
            subprocess.run(
                [
                    get_unwrapped_command(name="pip"),
                    "config",
                    "unset",
                    "global.index-url",
                ],
                capture_output=True,
            )
        except Exception:
            console.print("Failed to reset PIP global settings.")

    @classmethod
    def index_credentials(cls, ctx: typer.Context):
        api_key = None
        token = None

        if auth := getattr(ctx.obj, "auth", None):
            client = auth.client
            token = client.token.get("access_token") if client.token else None
            api_key = client.api_key

        auth_envelop = json.dumps(
            {
                "version": "1.0",
                "access_token": token,
                "api_key": api_key,
                "project_id": ctx.obj.project.id if ctx.obj.project else None,
            }
        )
        return base64.urlsafe_b64encode(auth_envelop.encode("utf-8")).decode("utf-8")

    @classmethod
    def default_index_url(cls) -> str:
        return "https://pypi.org/simple/"

    @classmethod
    def build_index_url(cls, ctx: typer.Context, index_url: Optional[str]) -> str:
        if index_url is None:
            index_url = PUBLIC_REPOSITORY_URL

        url = urlsplit(index_url)

        encoded_auth = cls.index_credentials(ctx)
        netloc = f"user:{encoded_auth}@{url.netloc}"

        if type(url.netloc) is bytes:
            url = url._replace(netloc=netloc.encode("utf-8"))
        elif type(url.netloc) is str:
            url = url._replace(netloc=netloc)

        return urlunsplit(url)



import platform
import time
from safety.console import main_console as console


def load_emoji(emoji):
    """
    Return the most appropriate emoji symbol based on terminal environment.

    Args:
        emoji (str): The emoji character to display, if supported

    Returns:
        str: The emoji if supported, or an appropriate alternative
    """
    # On non-Windows platforms, return the original emoji
    if platform.system() != "Windows":
        return emoji

    # Windows emoji mapping - add supported emojis or provide alternatives
    emoji_map = {
        "✓": "+",
        "🛡": "SHIELD",
        "⚠️": "!",
        "❌": "X",
        "ℹ️": "i",
        "🔒": "LOCK",
        "🔑": "KEY",
    }

    # Return the mapped version if it exists, otherwise return original
    # This allows Windows to display emojis it actually supports
    return emoji_map.get(emoji, emoji)


def typed_print(text, delay=0.02, console=console, style="bold", end_line=True):
    for char in text:
        console.print(char, end="", style=style)
        if char != "\n":
            time.sleep(delay)
    if end_line:
        console.line()


def progressive_print(sections, pause_between=0.7):
    for section in sections:
        console.print(section)
        time.sleep(pause_between)


def render_header(title, emoji="🛡", margin_left=0, margin_right=2):
    """
    Create a modern header with emoji that works cross-platform
    """
    header_text = f"{' ' * margin_left}{emoji}{title}{' ' * margin_right}"
    underline = (
        f"[blue]{'━' * (margin_left + len(emoji) + len(title) + margin_right)}[/blue]"
    )

    console.print()
    typed_print(header_text, style="bold white", delay=0.01, console=console)
    console.print(underline)
    console.print()
